#![cfg_attr(rustfmt, rustfmt_skip)]

pub static APP_ID: &str = "de.leopoldluley.Clapgrep";
pub static APP_PATH: &str = "/de/leopoldluley/Clapgrep";
pub static APP_VERSION: &str = "26.02";
pub static GRESOURCES_BUNDLE: &[u8] = include_bytes!("/build/clapgrep/src/build/assets/resources.gresource");
